@@ -1,12 +1,24 @@
-/* Simple JSON implementation
+/* bsjson.c - Simple JSON implementation
  *
- * Copyright (C) 2014 Borislav Sapundzhiev
+ * Copyright (C) 2014, Borislav Sapundzhiev
  *
- * This program is free software: you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation, either version 2 of the License, or (at
- * your option) any later version.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
  *
+ *    * Redistributions of source code must retain the above copyright notice,
+ *      this list of conditions and the following disclaimer.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
+ * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
+ * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <stdio.h>
@@ -14,10 +26,24 @@
 #include <string.h>
 #include <assert.h>
 #include <errno.h>
+#define oom() break
+#include "utstring.h"
 #include "bsjson.h"
 
 #define JSON_STACK_SIZE 32
 
+#define isNullorEmpty(str) \
+    (str == NULL || strlen(str) == 0)
+
+#define utstring_addchr(s, c) \
+    do {\
+        UT_string tmp; \
+        utstring_init(&tmp); \
+        utstring_printf(&tmp, "%c", c); \
+        utstring_concat(s, &tmp); \
+        utstring_done(&tmp); \
+    } while(0)
+
 enum eElemType {
     JSON_OBJ_B, JSON_OBJ_E, JSON_ARR_B, JSON_ARR_E,
     JSON_COLON, JSON_COMMA, JSON_QUOTE, JSON_LEFT,
@@ -54,7 +80,7 @@ JsonNode * JsonNode_Create()
     node->m_name = NULL;
     node->m_parent = NULL;
     node->m_pairs = cpo_array_create(4 , sizeof(JsonPair));
-    node->m_childs =  cpo_array_create(4 , sizeof(JsonNode));
+    node->m_childs = cpo_array_create(4 , sizeof(JsonNode));
     return node;
 }
 
@@ -65,7 +91,7 @@ JsonNode * JsonNode_createChild(JsonNode * node, String name, int type)
     child->m_parent = node;
     child->m_name = (name != NULL) ? strdup(name) : NULL;
     child->m_pairs = cpo_array_create(4 , sizeof(JsonPair));
-    child->m_childs =  cpo_array_create(4 , sizeof(JsonNode));
+    child->m_childs = cpo_array_create(4 , sizeof(JsonNode));
     return child;
 }
 
@@ -82,8 +108,8 @@ JsonNode * JsonNode_createArray(JsonNode * node, String name)
 void JsonNode_setPair(JsonNode * node, const String key, const String value )
 {
     JsonPair *a = (JsonPair*)cpo_array_push( node->m_pairs );
-    a->key =  strdup(key);
-    a->value =  strdup(value);
+    a->key = strdup(key);
+    a->value = strdup(value);
 }
 
 static int JsonPair_comparer(const void *a, const void *b)
@@ -196,18 +222,19 @@ void JsonNode_deleteTree(JsonNode *root)
         root = NULL;
     }
 }
-
-String JsonNode_getJSON(JsonNode *node)
+/* return allocated UT_string */
+static UT_string *JsonNode_getJSON_UT(JsonNode *node, String last)
 {
     int i, nPairs, nChilds;
-    String JSON = NULL;
-    bsstr *buff = bsstr_create("");
+    UT_string *buff;
+    utstring_new(buff);
+    if (buff == NULL) return NULL;
 
     if (!isNullorEmpty(node->m_name)) {
-        bsstr_printf(buff, "\"%s\":", node->m_name);
+        utstring_printf(buff, "\"%s\":", node->m_name);
     }
 
-    bsstr_printf(buff, "%s\n",  JSON_IS_OBJ(node) ? "{" : "[");
+    utstring_printf(buff, "%s\n",  JSON_IS_OBJ(node) ? "{" : "[");
     nPairs = JsonNode_getPairCount(node);
     nChilds = JsonNode_getChildCount(node);
 
@@ -215,29 +242,31 @@ String JsonNode_getJSON(JsonNode *node)
         JsonPair *pair = JsonNode_getPair(node, i);
 
         if (JSON_IS_ARRAY(node)) {
-            bsstr_printf(buff, "\"%s\"", pair->key);
+            utstring_printf(buff, "\"%s\"", pair->key);
         } else {
-            bsstr_printf(buff, "\"%s\":\"%s\"", pair->key, pair->value);
+            utstring_printf(buff, "\"%s\":\"%s\"", pair->key, pair->value);
         }
 
-        bsstr_printf(buff, "%s\n", (i < nPairs -1 || nChilds > 0) ? "," : "");
+        utstring_printf(buff, "%s\n", (i < nPairs -1 || nChilds > 0) ? "," : "");
     }
 
     for (i = 0; i < nChilds; i++) {
         JsonNode* child = JsonNode_getChild(node, i);
-        String childJSON = JsonNode_getJSON(child);
-        bsstr_add(buff, childJSON);
-        if (i < nChilds -1) {
-            int len = bsstr_length(buff)-1;
-            strncpy(bsstr_get_bufref(buff) + len, ",", 1);
-            bsstr_addchr(buff, '\n');
+        UT_string *childJSON = JsonNode_getJSON_UT(child, i == nChilds - 1 ? "\n" : ",\n");
+        if (childJSON != NULL) {
+            utstring_concat(buff, childJSON);
+            utstring_free(childJSON);
         }
-        free(childJSON);
     }
 
-    bsstr_printf(buff, "%s\n",  JSON_IS_OBJ(node) ? "}" : "]");
-    JSON = bsstr_release(buff);
-    return JSON;
+    utstring_printf(buff, "%s%s",  JSON_IS_OBJ(node) ? "}" : "]", last);
+    return buff;
+}
+/* return allocated string */
+String JsonNode_getJSON(JsonNode *node)
+{
+    UT_string *buff = JsonNode_getJSON_UT(node, "\n");
+    return utstring_release(buff);
 }
 /********************************************************************************/
 /* Parse JSON                                                                   */
@@ -254,8 +283,8 @@ struct  ParserInternal { /*Jsonlexer*/
     int line;
     int quote_begin;
     int is_value;
-    bsstr *key;
-    bsstr *value;
+    UT_string *key;
+    UT_string *value;
     cpo_array_t stack;
     struct JsonParser *parser;
     void (*startElem)(struct JsonParser *, const String, int);
@@ -267,8 +296,8 @@ static void JsonParser_internalCreate(struct ParserInternal *pi)
 {
     pi->error = JSON_ERR_NONE;
     pi->line = 0;
-    pi->key = bsstr_create("");
-    pi->value = bsstr_create("");
+    utstring_new(pi->key);
+    utstring_new(pi->value);
     pi->stack.v = calloc(JSON_STACK_SIZE, sizeof(char*));
     pi->stack.num = 0;
     pi->stack.max = JSON_STACK_SIZE;
@@ -281,21 +310,24 @@ static void JsonParser_internalCreate(struct ParserInternal *pi)
 static void JsonParser_internalReset(struct ParserInternal *pi)
 {
     pi->quote_begin = pi->is_value = 0;
-    bsstr_clear(pi->key);
-    bsstr_clear(pi->value);
+    utstring_clear(pi->key);
+    utstring_clear(pi->value);
 }
 
 static void JsonParser_internalDelete(struct ParserInternal *pi)
 {
-    bsstr_delete(pi->key);
-    bsstr_delete(pi->value);
+    utstring_free(pi->key);
+    utstring_free(pi->value);
     free(pi->stack.v);
 }
 
-static int JsonParser_internalBeginObj(struct  ParserInternal *pi, enum eElemType elemType)
+static int JsonParser_internalBeginObj(struct ParserInternal *pi, enum eElemType elemType)
 {
-    char *name = bsstr_get_buf(pi->key);
+    char *name = utstring_body(pi->key);
     void *ptr = stack_push_back(&(pi)->stack);
+    pi->key->n = 64;
+    pi->key->d = (char*)calloc(64, 1);
+    pi->key->i = 0;
     ARR_VAL(ptr) = ARR_VAL2PTR(name);
     if (elemType != JSON_ARR_B && elemType != JSON_OBJ_B) {
         pi->error = JSNON_ERR_NOTOBJ;
@@ -309,7 +341,7 @@ static int JsonParser_internalBeginObj(struct  ParserInternal *pi, enum eElemTyp
     return JSON_OK;
 }
 
-static int JsonParser_internalEndObj(struct  ParserInternal *pi, enum eElemType elemType)
+static int JsonParser_internalEndObj(struct ParserInternal *pi, enum eElemType elemType)
 {
     char *name;
     void *ptr = stack_pop_back(&(pi)->stack);
@@ -329,17 +361,17 @@ static int JsonParser_internalEndObj(struct  ParserInternal *pi, enum eElemType
     return 0;
 }
 
-static int JsonParser_internalData(struct  ParserInternal *pi)
+static int JsonParser_internalData(struct ParserInternal *pi)
 {
     if (pi->quote_begin) {
         pi->error = JSON_ERR_QUOTE;
         return JSON_NOK;
     }
 
-    if (bsstr_length(pi->key) || bsstr_length(pi->value)) {
+    if (utstring_len(pi->key) > 0 || utstring_len(pi->value) > 0) {
 
         if (pi->elemData) {
-            pi->elemData(pi->parser,  bsstr_get_bufref(pi->key), bsstr_get_bufref(pi->value) );
+            pi->elemData(pi->parser, utstring_body(pi->key), utstring_body(pi->value) );
         }
     }
 
@@ -383,7 +415,7 @@ static char JsonParser_prev_char(const char *p , int pos)
     (JsonParser_peek_char(p,pos) == Json_elem(JSON_CR)\
     || JsonParser_peek_char(p, pos) == Json_elem(JSON_LF))
 
-static int JsonParser_internalParse(struct  ParserInternal *pi, const char* json , int len)
+static int JsonParser_internalParse(struct ParserInternal *pi, const char* json , int len)
 {
     int i;
     char ch;
@@ -412,12 +444,18 @@ static int JsonParser_internalParse(struct  ParserInternal *pi, const char* json
             break;
 
         case JSON_QUOTE:
-            pi->quote_begin = !pi->quote_begin;
+            /* escaped quote in value */
+            if(JsonParser_prev_char(p, i) == Json_elem(JSON_LEFT)) {
+                UT_string *data = (!pi->is_value) ? pi->key : pi->value;
+                utstring_addchr(data, Json_elem(JSON_QUOTE));
+                break;
+            }
 
-            if( pi->quote_begin) {
+            pi->quote_begin = !pi->quote_begin;
+            if(pi->quote_begin) {
                 char prev = JsonParser_prev_char(p, i);
-                if(prev != Json_elem(JSON_COMMA) && prev !=  Json_elem(JSON_COLON)
-                        && prev != Json_elem(JSON_OBJ_B) &&  prev != Json_elem(JSON_ARR_B) ) {
+                if(prev != Json_elem(JSON_COMMA) && prev != Json_elem(JSON_COLON)
+                        && prev != Json_elem(JSON_OBJ_B) && prev != Json_elem(JSON_ARR_B)) {
                     pi->error = JSON_ERR_SYN;
                     break;
                 }
@@ -433,8 +471,8 @@ static int JsonParser_internalParse(struct  ParserInternal *pi, const char* json
                 pi->is_value = 1;
             } else {
                 /* colon in value */
-                bsstr *data = (!pi->is_value) ? pi->key : pi->value;
-                bsstr_addchr(data, Json_elem(JSON_COLON));
+                UT_string *data = (!pi->is_value) ? pi->key : pi->value;
+                utstring_addchr(data, Json_elem(JSON_COLON));
             }
             break;
         case JSON_COMMA:
@@ -442,8 +480,8 @@ static int JsonParser_internalParse(struct  ParserInternal *pi, const char* json
                 JsonParser_internalData(pi);
             } else {
                 /* comma in value */
-                bsstr *data = (!pi->is_value) ? pi->key : pi->value;
-                bsstr_addchr(data, Json_elem(JSON_COMMA));
+                UT_string *data = (!pi->is_value) ? pi->key : pi->value;
+                utstring_addchr(data, Json_elem(JSON_COMMA));
             }
             break;
 
@@ -464,9 +502,9 @@ static int JsonParser_internalParse(struct  ParserInternal *pi, const char* json
         case JSON_HEX:  /*TODO:*/
         case JSON_INVALID:
             if (pi->quote_begin && !pi->is_value) {
-                bsstr_addchr(pi->key, ch);
+                utstring_addchr(pi->key, ch);
             } else if (pi->quote_begin && pi->is_value) {
-                bsstr_addchr(pi->value, ch);
+                utstring_addchr(pi->value, ch);
             } else {
                 //printf("[skiped] '%c' [0x%x]\n", ch,ch);
             }
@@ -568,14 +606,19 @@ static void JsonParser_stripCommentsFromBuffer(char *buff, long size)
 {
     long i;
     for(i = 0; i < size; i++) {
-        if((buff[i] == '/' && buff[i+1] == '/') || buff[i] == '#' || buff[i] == '"') {
+        if(((i < size - 1) && buff[i] == '/' && buff[i+1] == '/') || buff[i] == '#' || buff[i] == '"') {
             char s = buff[i];
-            while(buff[i] != '\n' && buff[i] != 0) {
+            long j = i;
+            do {
                 if(s != '"') {
                     buff[i] = ' ';
                 }
+                else if (i > j && buff[i] == '"') {
+                    i++;
+                    break;
+                }
                 i++;
-            }
+            } while(buff[i] != '\n' && buff[i] != 0);
         }
     }
 }
@@ -583,23 +626,28 @@ static void JsonParser_stripCommentsFromBuffer(char *buff, long size)
 JsonNode * JsonParser_parseFile(struct JsonParser *parser, const char * fileName)
 {
     char * buffer = 0;
-    long length =0;
+    long length = 0, read = 0;
     JsonNode * root = NULL;
-    FILE *f = fopen (fileName, "r");
+    FILE *f = fopen (fileName, "rb");
 
-    if (f) {
+    if (f != NULL) {
         fseek (f, 0, SEEK_END);
         length = ftell (f);
         fseek (f, 0, SEEK_SET);
         buffer = (char*) malloc (length + 1);
-        if (buffer) {
-            fread (buffer, sizeof(char), length, f);
-            buffer[length] = '\0';
+        if (buffer != NULL) {
+            read = fread (buffer, sizeof(char), length, f);
+            buffer[read] = '\0';
         }
-
         fclose (f);
-        JsonParser_stripCommentsFromBuffer(buffer, length);
-        root = JsonParser_parse(parser,  buffer);
+        if (read == length) {
+            JsonParser_stripCommentsFromBuffer(buffer, length);
+            root = JsonParser_parse(parser,  buffer);
+        } else {
+            parser->m_errorString = strerror(errno);
+            parser->m_errorLine = 0;
+            parser->m_errorLineSet = 0;
+        }
         free(buffer);
     } else {
         parser->m_errorString = strerror(errno);
